/*	$NetBSD$	*/

/*
 * Copyright (c) 2001 ITOH Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/endian.h>
#include <time.h>
#include <unistd.h>
#include <strings.h>
#include <stdlib.h>
#include <err.h>

#include "dcmem.h"

void
dcmem_start_write(m)
	struct dcmem *m;
{
	sigset_t set;

	if (m->dm_writing == 0) {
		m->dm_writing = 1;
		sigfillset(&set);
		sigprocmask(SIG_BLOCK, &set, &m->dm_osigset);
	}
}

void
dcmem_end_write(m)
	struct dcmem *m;
{

	if (m->dm_writing) {
		m->dm_writing = 0;
		sigprocmask(SIG_SETMASK, &m->dm_osigset, (sigset_t *) 0);
	}
}

int
dcmem_write_1(m, ptr, blk)
	struct dcmem *m;
	const void *ptr;
	int blk;
{
	u_int8_t buf[DCMEM_BLK_SZ];
	int retrycnt = DCMEM_RETRYCNT;
	off_t off = blk * DCMEM_BLK_SZ;
	int fd = m->dm_fd;
	int r, s = DCMEM_EHWPROBLEM;

	for (;;) {
		if (retrycnt-- == 0)
			break;

		/*
		 * read, compare, write if differs
		 */
		if ((r = dcmem_read_1(m, buf, blk)) != 0)
			return r;	/* read failed */

		if (memcmp(ptr, buf, DCMEM_BLK_SZ) == 0)
			return 0;

		/* read, compared, and differs */
		s = DCMEM_EHWPROBLEM;

		if (m->dm_flags & DCMEM_FRDONLY)
			return DCMEM_ERDONLY;

		dcmem_start_write(m);

		if ((r = pwrite(fd, ptr, DCMEM_BLK_SZ, off)) < 0) {
			s = DCMEM_ESYS;
			goto retry;
		} else if (r != DCMEM_BLK_SZ) {
			s = DCMEM_EUXEOF;
			goto retry;
		}

		continue;

	retry:
#ifdef DCMEM_DEBUG
		warn("dcmem_write_1: retrying: blk %u, err %d", blk, s);
#endif
		usleep(100000);
	}

#ifdef DCMEM_DEBUG
	warn("dcmem_write_1: write failed: blk %u, err %d", blk, s);
#endif
	return s;
}

/*
 * write user data on the free blocks
 */
int
dcmem_write_data(m, ptr, sz, pfatstart)
	struct dcmem *m;
	const void *ptr;
	int sz;		/* size in block */
	int *pfatstart;
{
	int f, fst, prevf, nfr;
	int i;
	const u_int8_t *p;
	int s;
	int last_broken;

	if (sz <= 0)
		return 0;	/* no work to do */
	nfr = dcmem_fat_cnt_free_data(m);
	if (nfr < sz)
		return DCMEM_ENOSPC;

	fst = -1;
	prevf = -1;
#if 0	/* XXX shut up gcc */
	f = -1;
#endif
	for (i = 0, p = ptr; i < sz; i++, p += DCMEM_BLK_SZ) {
		last_broken = DCMEM_FAT_END;
	again:
		f = dcmem_fat_find_free_data(m);

		/* this check is redundant */
		if (f == -1)
			return DCMEM_ENOSPC;

		if ((s = dcmem_write_1(m, p, f)) == DCMEM_EHWPROBLEM) {
			/*
			 * data could not be written
			 */
			nfr--;
			if (nfr < sz) {
				/* no space for recovery */
#ifdef DCMEM_DEBUG
				warn("dcmem_write_data: no space for recovery");
#endif
				s = DCMEM_ENOSPC;
				goto cleanup;
			}
			/*
			 * Temporarily put the broken block in a chain
			 * so that it won't be used to have this data.
			 */
			DCMEM_FAT_MARK(m, f, last_broken);
			last_broken = f;
			goto again;
		} else if (s) {
			/*
			 * I/O error
			 * cleanup may be useless...
			 */
		cleanup:
			/* free partial chain */
			if (fst != -1) {
				DCMEM_FAT_MARK(m, f, DCMEM_FAT_END);
				dcmem_fat_free_chain_data(m, fst);
			}
			/* free ``broken'' blocks */
			if (last_broken != DCMEM_FAT_END)
				dcmem_fat_free_chain_data(m, last_broken);
			return s;
		}

		/* make fat chain */
		if (prevf == -1) {
			fst = f;
		} else {
			DCMEM_FAT_MARK(m, prevf, f);
		}
		DCMEM_FAT_MARK(m, f, DCMEM_FAT_END);
		prevf = f;

		/*
		 * The ``broken'' blocks cannot have this data
		 * but it may possibly have other bit patterns,
		 * so mark it as free rather than as broken.
		 */
		if (last_broken != DCMEM_FAT_END)
			dcmem_fat_free_chain_data(m, last_broken);
	}

	*pfatstart = fst;

	return 0;
}

/*
 * write program at the top of VMU
 */
int
dcmem_write_program(m, ptr, sz)
	struct dcmem *m;
	const void *ptr;
	int sz;		/* size in block */
{
	int f, nfr;
	int i;
	const u_int8_t *p;
	int s;

	if (sz <= 0)
		return 0;	/* no work to do */
	nfr = dcmem_fat_cnt_free_program(m);
	if (nfr < sz)
		return DCMEM_ENOSPC;

	f = DCMEM_START_PROGRAM - 1;
	for (i = 0, p = ptr; i < sz; i++, p += DCMEM_BLK_SZ) {
		f++;

		if ((s = dcmem_write_1(m, p, f)) != 0) {
#ifdef DCMEM_DEBUG
			if (s == DCMEM_EHWPROBLEM) {
				/*
				 * data could not be written
				 */
				/* impossible to recover */
				warn("dcmem_write_program: write failed");
			}
#endif
			DCMEM_FAT_MARK(m, f, DCMEM_FAT_END);
			dcmem_fat_free_chain_data(m, DCMEM_START_PROGRAM);
			return s;
		}

		if (f != DCMEM_START_PROGRAM) {
			DCMEM_FAT_MARK(m, f - 1, f);
		}
	}

	DCMEM_FAT_MARK(m, f, DCMEM_FAT_END);

	return 0;
}

/*
 * overwrite system files on the current FAT chain
 */
int
dcmem_update_sys(m, ptr, isfat)
	struct dcmem *m;
	void *ptr;
	int isfat;	/* FAT / dir */
{
	int f, fst, prevf, nfr, fnew;
	int i;
	u_int8_t *p;
	int sz;
	int s;

	nfr = dcmem_fat_cnt_free_sys(m);

	if (isfat) {
		sz = m->dm_fatsz;
		f = m->dm_fatloc;
#if 1 /* XXX for now */
		if (sz != 1)
			return DCMEM_EINCONSISTENT;
#endif
	} else {
		sz = m->dm_dirsz;
		f = m->dm_dirloc;
	}

	fst = -1;
	prevf = -1;
	for (i = 0, p = ptr; i < sz; i++, p += DCMEM_BLK_SZ) {
		/* for recovery */
		if (f == -1)
			f = dcmem_fat_find_free_sys(m);
	again:
		/* this shall not happen */
		if (f >= m->dm_maxblk || f < m->dm_minblk + m->dm_nblk)
			return DCMEM_EINCONSISTENT;

		if ((s = dcmem_write_1(m, p, f)) == DCMEM_EHWPROBLEM) {
			/*
			 * data could not be written
			 */
			nfr--;
			if (nfr < sz) {
				/* No space left for recovery. */
#ifdef DCMEM_DEBUG
				warn("dcmem_update_sys: Cry, for \"VMU\" the broken");
#endif
				return s;
			}
			DCMEM_FAT_MARK(m, f, DCMEM_FAT_BROKEN);

			/*
			 * If the top block is broken, assign the top to
			 * a fresh block in order to reduce # times of
			 * rewriting the root block.
			 */
			if (prevf == -1) {
				/* top blk */
				fnew = dcmem_fat_find_free_sys(m);
				DCMEM_FAT_MARK(m, fnew, f);
				f = fnew;
			} else {
				f = dcmem_fat_next(m, f);
				if (f == -1)
					f = dcmem_fat_find_free_sys(m);
			}
			goto again;
		} else if (s) {
			return s;
		}

		/* make fat chain */
		if (prevf == -1) {
			fst = f;
		} else {
			DCMEM_FAT_MARK(m, prevf, f);
		}
		prevf = f;
		f = dcmem_fat_next(m, f);
		DCMEM_FAT_MARK(m, prevf, DCMEM_FAT_END);
	}

	/*
	 * Update root block if the top location of the system file is changed
	 */
	if (isfat) {
		if (fst != m->dm_fatloc) {
			m->dm_fatloc = fst;
			m->dm_rblk.dr_fatloc = htole16(fst);
			m->dm_rblk_changed = 1;
		}
	} else {
		if (fst != m->dm_dirloc) {
			m->dm_dirloc = fst;
			m->dm_rblk.dr_dirloc = htole16(fst);
			m->dm_rblk_changed = 1;
		}
	}

	return 0;
}

/*
 * emergency save
 * write system blocks to a file
 */
int
dcmem_emergency_save(m)
	struct dcmem *m;
{
	int fd;
	char fn[sizeof DCMEM_SAVE_TEMPLATE];
	int st;
	int sysstart;
	char *p;
	int f;

	st = 0;

	strcpy(fn, DCMEM_SAVE_TEMPLATE);
	if ((fd = mkstemp(fn)) < 0) {
		warn("emergency save failed");
		return DCMEM_ESYS;
	}

	sysstart = m->dm_minblk + m->dm_nblk;

	/* save directory */
	p = (void *) m->dm_dir;
	for (f = m->dm_dirloc; f != -1; f = dcmem_fat_next(m, f)) {
		if (pwrite(fd, p, DCMEM_BLK_SZ,
		    (off_t) ((f - sysstart) * DCMEM_BLK_SZ)) != DCMEM_BLK_SZ)
			st = DCMEM_ESYS;
		p += DCMEM_BLK_SZ;
	}

	/* save FAT */	/* XXX assume m->dm_fatsz == 1 */
	if (pwrite(fd, m->dm_fat, DCMEM_BLK_SZ,
	    (off_t) ((m->dm_fatloc - sysstart) * DCMEM_BLK_SZ)) != DCMEM_BLK_SZ)
		st = DCMEM_ESYS;

	/* save root */
	if (pwrite(fd, &m->dm_rblk, DCMEM_BLK_SZ,
	    (off_t) ((m->dm_rbloc - sysstart) * DCMEM_BLK_SZ)) != DCMEM_BLK_SZ)
		st = DCMEM_ESYS;

	if (close(fd))
		st = DCMEM_ESYS;

	if (st == 0) {
		warnx("emergency save at %s\n\
how to recover:\n\
  (dd if=<dev_name> bs=%d count=%d; cat %s) >recoverd.dcm",
		    fn, DCMEM_BLK_SZ, sysstart, fn);
	} else {
		warn("emergency save failed");
		unlink(fn);
	}

	return st;
}

/*
 * update cached directory / FAT / root block
 */
int
dcmem_flush(m)
	struct dcmem *m;
{
	int st;

	/* update directory */
	if ((st = dcmem_update_sys(m, m->dm_dir, 0)) != 0) {
		/* failed --- updating FAT/root does not make sense */
		goto out;
	}

	/* update FAT */
	if ((st = dcmem_update_sys(m, m->dm_fat, 1)) != 0) {
		/* failed --- updating root does not make sense */
		goto out;
	}

	/* update root block */
	if (m->dm_rblk_changed) {
#ifdef DCMEM_DEBUG
		warn("dcmem_flush: updating root block");
#endif
		if ((st = dcmem_write_1(m, &m->dm_rblk, m->dm_rbloc)) == 0)
			m->dm_rblk_changed = 0;
	}

out:
	if (st && st != DCMEM_ERDONLY) {
		warnx("fatal: failed to sync system blocks");
		dcmem_emergency_save(m);
	}

	/* all written */
	dcmem_end_write(m);

	return st;
}
