/*	$NetBSD$	*/

/*
 * Copyright (c) 2001 ITOH Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/endian.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <err.h>
#include <stdio.h>

#include "dcmem.h"

int main __P((int, char *[]));

long dcmembuf[(DCMEM_BLK_SZ * DCMEM_SIZE + 512) / sizeof(long)];

/* ARGSUSED */
int
cmd_fread(m, ac, av, opts, oargs)
	struct dcmem *m;
	int ac;
	char *av[];
	int opts[];
	char *oargs[];
{
	struct dcdirent *de = (void *) dcmembuf;
	int st;
	size_t nbyte;
	ssize_t nwr;
	int fd;

	/*
	 * read file from flash
	 */
	if ((st = dcmem_read_file_name(m, (u_int8_t *) av[0],
	    (u_int8_t *) dcmembuf + sizeof(struct dcdirent), de)) != 0) {
		warnx("fread: %s: %s", av[0], dcmem_strerror(st));
		return st;
	}

	/*
	 * don't copy copy-protected file
	 */
	if (de->dd_copyprot == DCMEM_COPY_PROTECTED
	    && !opts[0]		/* -d */
	    && !opts[1]) {	/* -f */
		warnx("fread: %s: copy protected -- use fread -d (delete from flash after read)", av[0]);
		return 1;
	}

	/*
	 * write
	 */
	nbyte = sizeof(struct dcdirent) + DCMEM_BLK_SZ * le16toh(de->dd_size);

	if (av[1][0] == '-' && av[1][1] == '\0') {
		fd = STDOUT_FILENO;
	} else {
		if ((fd = open(av[1], O_CREAT|O_TRUNC|O_WRONLY, 0666)) < 0) {
			warn("fread: %s", av[1]);
			return -1;
		}
	}
	if ((size_t) (nwr = write(fd, dcmembuf, nbyte)) != nbyte) {
		if (nwr < 0)
			warn("fread: %s: write", av[1]);
		else
			warnx("fread: %s: short write", av[1]);
		if (fd != STDOUT_FILENO)
			close(fd);
		return -1;
	}

	/*
	 * make sure the file is written if we are removing it from flash
	 */
	if (opts[0]) {	/* -d */
		if (fsync(fd)) {
			warn("fread: %s: fsync", av[1]);
			return -1;
		}
	}
	if (fd != STDOUT_FILENO && close(fd)) {
		warn("fread: %s: close", av[1]);
		return -1;
	}

	/*
	 * remove from flash (-d flag)
	 */
	if (opts[0])	/* -d */
		(void) dcmem_dir_remove_name(m, (u_int8_t *) av[0]);

	return 0;
}

/* ARGSUSED */
int
cmd_fwrite(m, ac, av, opts, oargs)
	struct dcmem *m;
	int ac;
	char *av[];
	int opts[];
	char *oargs[];
{
	int fd;
	ssize_t nrd;
	struct dcdirent *de = (void *) dcmembuf;
	int st;

	/*
	 * read
	 */
	if (av[0][0] == '-' && av[0][1] == '\0') {
		fd = STDIN_FILENO;
	} else {
		if ((fd = open(av[0], O_RDONLY)) < 0) {
			warn("fwrite: %s", av[0]);
			return -1;
		}
	}
	nrd = read(fd, dcmembuf, sizeof dcmembuf);
	if (fd != STDIN_FILENO)
		close(fd);

	if (nrd < 0) {
		warn("fwrite: %s: read", av[0]);
		return -1;
	} else if (nrd < sizeof(struct dcdirent)) {
		warnx("fwrite: %s: short read", av[0]);
		return -1;
	} else if (nrd !=
	    sizeof(struct dcdirent) + DCMEM_BLK_SZ * le16toh(de->dd_size)) {
		warnx("fwrite: %s: invalid file format", av[0]);
		return -1;
	}

	/* -f: overwrite */
	if (opts[0])	/* -f */
		(void) dcmem_dir_remove_name(m, de->dd_name);

	/*
	 * write to flash
	 */
	if ((st = dcmem_write_file_dirent(m, de,
	    (u_int8_t *) dcmembuf + sizeof(struct dcdirent))) != 0) {
		warnx("fwrite: %.*s: %s",
		    (int) sizeof de->dd_name, de->dd_name, dcmem_strerror(st));
		return st;
	}

	return 0;
}

static int
bcd(num)
	int num;
{

	return (num / 10) << 4 | num % 10;
}

/* ARGSUSED */
int
cmd_vmiwrite(m, ac, av, opts, oargs)
	struct dcmem *m;
	int ac;
	char *av[];
	int opts[];
	char *oargs[];
{
	int fd;
	ssize_t nrd;
	struct dcvmi *vmi;
	struct dcdirent dir;
	int st;
	int i;
	u_int32_t nbyte;
	unsigned padsz;

	/*
	 * read VMI file
	 */
	if ((fd = open(av[0], O_RDONLY)) < 0) {
		warn("vmiwrite: %s", av[0]);
		return -1;
	}
	nrd = read(fd, dcmembuf, sizeof dcmembuf);
	close(fd);

	if (nrd < 0) {
		warn("vmiwrite: %s: read", av[0]);
		return -1;
	} else if (nrd != sizeof(struct dcvmi)) {
		warnx("vmiwrite: %s: invalid file size", av[0]);
		return 1;
	}
	vmi = (void *) dcmembuf;

	/*
	 * check VMI and create the directory entry
	 */
#if 0
	printf("%x %x  %x %x  %x %x  %x %x\n",
	    vmi->di_chkbyte[0], (vmi->di_resource[0] & DCMEM_VMI_CHKAND0),
	    vmi->di_chkbyte[1], (vmi->di_resource[1] & DCMEM_VMI_CHKAND1),
	    vmi->di_chkbyte[2], (vmi->di_resource[2] & DCMEM_VMI_CHKAND2),
	    vmi->di_chkbyte[3], (vmi->di_resource[3] & DCMEM_VMI_CHKAND3));
#endif
#ifndef __GNUC__
	if (vmi->di_chkbyte[0] != (vmi->di_resource[0] & DCMEM_VMI_CHKAND0) ||
	    vmi->di_chkbyte[1] != (vmi->di_resource[1] & DCMEM_VMI_CHKAND1) ||
	    vmi->di_chkbyte[2] != (vmi->di_resource[2] & DCMEM_VMI_CHKAND2) ||
	    vmi->di_chkbyte[3] != (vmi->di_resource[3] & DCMEM_VMI_CHKAND3)) {
		warnx("vmiwrite: %s: invalid file format", av[0]);
		return 1;
	}
#else	/* compiler bug workaround */
	if (vmi->di_chkbyte[0] != (vmi->di_resource[0] & DCMEM_VMI_CHKAND0))
		goto invfmt;
	if (vmi->di_chkbyte[1] != (vmi->di_resource[1] & DCMEM_VMI_CHKAND1))
		goto invfmt;
	if (vmi->di_chkbyte[2] != (vmi->di_resource[2] & DCMEM_VMI_CHKAND2))
		goto invfmt;
	if (vmi->di_chkbyte[3] != (vmi->di_resource[3] & DCMEM_VMI_CHKAND3)) {
	invfmt:
		warnx("vmiwrite: %s: invalid file format", av[0]);
		return 1;
	}
#endif
	if (/* ??? vmi->di_ver != htole16(DCMEM_VMI_VERSION) || */
	    vmi->di_filenum != htole16(DCMEM_VMI_FILENUM)) {
		warnx("vmiwrite: %s: unsupported file format", av[0]);
		return 1;
	}

	nbyte = le32toh(vmi->di_size);
	if (nbyte > m->dm_nblk * DCMEM_BLK_SZ) {
		warnx("vmiwrite: %s: requests too large VMS file", av[0]);
		return 1;
	}

	memset(&dir, 0, sizeof dir);
	dir.dd_type = vmi->di_mode & htole16(DCMEM_VMI_MODE_PROGRAM) ?
	    DCMEM_FTYPE_PROGRAM : DCMEM_FTYPE_DATA;
	dir.dd_copyprot = vmi->di_mode & htole16(DCMEM_VMI_MODE_COPY_PROTECTED)?
	    DCMEM_COPY_PROTECTED : DCMEM_COPY_OK;
	memcpy(dir.dd_name, vmi->di_fn, sizeof dir.dd_name);
	i = le16toh(vmi->di_year);
	dir.dd_timestamp[0] = bcd(i / 100);
	dir.dd_timestamp[1] = bcd(i % 100);
	dir.dd_timestamp[2] = bcd(vmi->di_month);
	dir.dd_timestamp[3] = bcd(vmi->di_day);
	dir.dd_timestamp[4] = bcd(vmi->di_hour);
	dir.dd_timestamp[5] = bcd(vmi->di_minute);
	dir.dd_timestamp[6] = bcd(vmi->di_second);
	dir.dd_timestamp[7] = vmi->di_weekday;
	dir.dd_size = htole16((nbyte + DCMEM_BLK_SZ - 1) / DCMEM_BLK_SZ);
	dir.dd_hdroff = vmi->di_mode & htole16(DCMEM_VMI_MODE_PROGRAM) ?
	    htole16(1) : 0;

#if 1	/* fool proof */
	vmi = 0;
#endif
	/*
	 * read VMS file
	 */
	if ((fd = open(av[1], O_RDONLY)) < 0) {
		warn("vmiwrite: %s", av[1]);
		return -1;
	}
	nrd = read(fd, dcmembuf, sizeof dcmembuf);
	close(fd);

	if (nrd < 0) {
		warn("vmiwrite: %s: read", av[1]);
		return -1;
	} else if (nrd != (ssize_t) nbyte) {
		warnx("vmiwrite: %s: does not match the VMI file", av[1]);
		return -1;
	}

	/* clear padding just in case */
	padsz = (-(unsigned)nbyte) & (DCMEM_BLK_SZ - 1);
	if (padsz)
		memset((char *) dcmembuf + nbyte, 0, padsz);

	/* -f: overwrite */
	if (opts[0])	/* -f */
		(void) dcmem_dir_remove_name(m, dir.dd_name);

	/*
	 * write to flash
	 */
	if ((st = dcmem_write_file_dirent(m, &dir, dcmembuf)) != 0) {
		warnx("vmiwrite: %.*s: %s",
		    (int) sizeof dir.dd_name, dir.dd_name, dcmem_strerror(st));
		return st;
	}

	return 0;
}

/* ARGSUSED */
int
cmd_rm(m, ac, av, opts, oargs)
	struct dcmem *m;
	int ac;
	char *av[];
	int opts[];
	char *oargs[];
{

	/*
	 * remove a file from flash
	 */
	return dcmem_dir_remove_name(m, (u_int8_t *) av[0]);
}

/* ARGSUSED */
int
cmd_ls(m, ac, av, opts, oargs)
	struct dcmem *m;
	int ac;
	char *av[];
	int opts[];
	char *oargs[];
{
	struct dcdirent *de;
	int pos;
	int need_hdr;
	static const char dow[] = "MonTueWedThuFriSatSun";

	printf("free: data %d program %d\n",
	    dcmem_fat_cnt_free_data(m), dcmem_fat_cnt_free_program(m));

	pos = 0;
	need_hdr = 1;
	while ((de = dcmem_dir_enumerate(m, &pos)) != 0) {
		if (need_hdr) {
			printf("type        size date[Y/M/D]      time      name\n");
			need_hdr = 0;
		}
		if (de->dd_type == DCMEM_FTYPE_DATA)
			printf("data");
		else if (de->dd_type == DCMEM_FTYPE_PROGRAM)
			printf("prog");
		else
			printf("0x%02x", de->dd_type);
		printf("%s %4d %04x/%02x/%02x ",
		    de->dd_copyprot == DCMEM_COPY_PROTECTED ?
			",nocopy" :  "       ",
		    de->dd_size,
		    (de->dd_timestamp[0] << 8) | de->dd_timestamp[1],
		    de->dd_timestamp[2], de->dd_timestamp[3]);
		if (de->dd_timestamp[7] <= 7)
			printf("(%.3s)", dow + de->dd_timestamp[7] * 3);
		else
			printf("0x%02x ", de->dd_timestamp[7]);
		printf(" %02x:%02x:%02x  %.*s\n",
		    de->dd_timestamp[4], de->dd_timestamp[5],
		    de->dd_timestamp[6],
		    (int) sizeof de->dd_name, de->dd_name);
	}
	if (need_hdr)
		printf("no files\n");

	return 0;
}

const struct cmd {
	const char *name;
	const char *optstr;
	const char *helpstr;
	int (*func) __P((struct dcmem *, int, char **, int *, char **));
	int narg;
} cmdlist[] = {
#if 0
	{ "read", "h", "read [-h] <vms_filename> <file>", 2 },
	{ "write", "h", "write [-h] data|prog[,nocopy] <vms_filename> <file>", 2 },
#endif
	{ "fread", "df", "fread [-df] <vms_filename> <file>", cmd_fread, 2 },
	{ "fwrite", "f", "fwrite [-f] <file>", cmd_fwrite, 1 },
	{ "vmiwrite", "f", "vmiwrite [-f] <file.VMI> <file.VMS>", cmd_vmiwrite, 2 },
	{ "rm", "", "rm <vms_filename>", cmd_rm, 1 },
	{ "ls", "", "ls", cmd_ls, 0 },
	{ 0, 0, 0, 0, 0 }
};

#define MAXOPTS	10

static void
usage()
{

	/* XXX */
	fprintf(stderr, "syntax error\n");
	exit(1);
}

static const struct cmd *
lookupcmd(name)
	const char *name;
{
	const struct cmd *c, *cpartial;
	size_t len;
	int npartial;

	len = strlen(name);
	npartial = 0;
#if 1	/* XXX shut up gcc */
	cpartial = 0;
#endif
	for (c = cmdlist; c->name; c++) {
		if (strncmp(name, c->name, len) == 0) {
			if (c->name[len] == '\0')
				return c;	/* exact match */
			/* partial match */
			cpartial = c;
			npartial++;
		}
	}

	if (npartial == 1)
		return cpartial;

	return 0;
}

/*
 * if parseonly
 *	detect systax errors
 *	do not report warnings
 *	do not actually execute commands
 * else
 *	execute commands
 *	report non-fatal syntactical warnings
 */
int
parse_args(m, ac, av, parseonly)
	struct dcmem *m;
	int ac;
	char *av[];
	int parseonly;
{
	const struct cmd *cmd;
	int ch, i, opt_found;
	int opts[MAXOPTS];
	char *oargs[MAXOPTS];
	const char *p;
	int st = 0;

	while (ac > 0) {
		if ((cmd = lookupcmd(av[0])) == 0) {
			warnx("%s: unknown command", av[0]);
			usage();
		}
		for (i = 0; i < MAXOPTS; i++) {
			opts[i] = 0;
			oargs[i] = 0;
		}
		optind = 1;
		while ((ch = getopt(ac, av, cmd->optstr)) != -1) {
			opt_found = 0;
			for (p = cmd->optstr, i = 0; *p; p++, i++) {
				if (*p == ch) {
					opt_found = 1;
					opts[i] = 1;
					if (p[1] == ':') {
						if (oargs[i]) {
							if (!parseonly)
								warnx(
			"%s: multiple -%c flags (only the first one is valid)",
								    cmd->name,
								    ch);
						} else
							oargs[i] = optarg;
					}
				}
				if (p[1] == ':')
					p++;
			}
			if (opt_found == 0) {
				warnx("bad option: usage: %s", cmd->helpstr);
				usage();
			}
		}
		ac -= optind;
		av += optind;
		if (ac < cmd->narg) {
			warnx("too few arguments: usage: %s", cmd->helpstr);
			usage();
		}
#if 0
fprintf(stderr, "command: %s, %d, %d\n", cmd->name, optind, ac);
#endif
		if (!parseonly)
			st |= (*cmd->func)(m, ac, av, opts, oargs);
		ac -= cmd->narg;
		av += cmd->narg;
	}

	return st;
}

int
main(argc, argv)
	int argc;
	char *argv[];
{
	int ch;
	int nowrite = 0;
	const char *devname = 0;
	int ac;
	char **av;
	struct dcmem *m;
	int st;

	while ((ch = getopt(argc, argv, "f:n")) != -1)
		switch (ch) {
		case 'n':
			nowrite = 1;
			break;
		case 'f':
			devname = optarg;
			break;
		case '?':
		default:
			/* XXX */
			usage();
		}

	ac = argc -= optind;
	av = argv += optind;

/*	if (ac == 0) cmdline();	*/

	if (parse_args((struct dcmem *) 0, ac, av, 1))
		return 1;
#define DEFAULT_DEV	"/dev/mmem0"
	if (devname == 0)
		devname = DEFAULT_DEV;

	if ((m = dcmem_open(devname, nowrite ? DCMEM_FRDONLY : 0, &st)) == 0) {
		errx(1, "%s: %s", devname, dcmem_strerror(st));
	}

	st = parse_args(m, ac, av, 0);

	if ((st = dcmem_close(m)) != 0) {
		errx(1, "%s: dcmem_close: %s", devname, dcmem_strerror(st));
	}

	return st;
}
